#!/usr/bin/perl -w

use strict;
use DBI();
use DBD::Pg();
use Getopt::Std;
use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );
my %opt;
getopts('d:n:c:l', \%opt);

if ( !( scalar %opt ) ) {
	print <<EOT;
Author: Teodor Sigaev <teodor\@sigaev.ru>
Usage:
$0 -d DATABASE [-n N] [-c N] [-l]
-d DATABASE  - DATABASE name
-n N	- number of rows
-c N    - number of flow
-l      - linear mode

EOT
	exit;
}

$opt{d} ||= 'qq';
my $db = $opt{d};

my $n = $opt{n};
$n>0 || die;
my $c = $opt{c};
$c>0 || die; 

print "Start: ",( ( $opt{l} ) ? 'single' : 'parallel')." mode";
if ( !$opt{l} ) {
	print " with $c flows";
}
print "\n";

my $t0 = [gettimeofday];
for(my $nf=0;$nf<$c;$nf++) {
	if ( $opt{l} ) {
		run_fill($n,$c,$nf);
	} elsif ( fork()==0 ) {
		#child
		sleep(1); 
		run_fill($n,$c,$nf);
		exit(0);
	}
}


my $status = 0;
if ( !$opt{l} ) {
	my $pid;
	while( ($pid=wait())>0 ) {
		$status ||= $?>>8;
	}
}

my $elapsed = tv_interval ( $t0, [gettimeofday]);
my $fmt = sprintf("elapsed time: %.02f sec", $elapsed); 
print "All flow finish; status: $status; $fmt\n";

exit( $status );

sub run_fill {
	my ($n, $c, $nf) = @_;

	srand($nf+1);
	
	my ($ni, $nu, $nd, $nv, $nfull, $nt) = (0,0,0,0,0,0); 

	my $dbi=DBI->connect('DBI:Pg:dbname='.$opt{d}.";port=5432") || die "Couldn't connect DB: $opt{d} !\n";
	$dbi->do("set enable_seqscan=off;") || die;
	$dbi->begin_work; $nt++;
	for(my $i=$nf;$i<$n;$i+=$c) {
		my $str=join(',', map { int(rand()*100) } 0..(int(70*rand()*rand()*rand()*rand()*rand())) );
		if ( rand() < 0.005 && length($str)<4 ) {
			$dbi->do("delete from wow where a \@ '{$str}' and n\%$c=$nf;") || die "DELETE"; 
			my $n = get_res($dbi, "select count(*) as check from wow where a \@ '{$str}' and n\%$c=$nf;");
			die "Problem with delete {$str}:$nf count:$n" if $n>0;
			$i-=$c;
			$nd++;
		} elsif ( rand()<0.05 && $str=~/,/ ) {
			$dbi->do("update wow set a = a || '{101}'::int[] where a \@ '{$str}' and n\%$c=$nf;") || die "UPDATE";
			my $n=get_res($dbi, "select count(*) as check from wow where a \@ '{$str}' and n\%$c=$nf and not a && '{101}';");
			die "Problem with update {$str}:$nf count:$n" if $n>0;
			$i-=$c;
			$nu++;
		} elsif ( rand()<0.0001 ) {
			$dbi->commit;
			my $full = ( rand()<0.2 ) ? 'full' : '';
			$nfull++ if length $full;
			$dbi->do("vacuum $full wow;") || die "VACUUM $full";
			$dbi->begin_work; $nt++;
			$i-=$c;
			$nv++;
		} elsif ( rand()<0.01 ) {
			$dbi->commit;
			$dbi->begin_work; $nt++;
			$i-=$c;
		} else { 
			$dbi->do("insert into wow values ($i, '{$str}');") || die "INSERT";
			$ni++;
		}
	}
	$dbi->commit;
	$dbi -> disconnect;

	print "$nf flow finish. Stats: ni:$ni nu:$nu nd:$nd nv:$nv(nf:$nfull) nt:$nt\n";
}

sub get_res { 
        my ($dbi, $sql) = @_;
        my $sth=$dbi->prepare($sql) || die "SQL: $sql";
        $sth->execute() || die "SQL: $sql";
        my $r = $sth->fetchrow_hashref;
        $sth->finish;
        return $r->{check};
}

